/* vi: set sw=4 ts=4: */
/*
 * Mini dd implementation for busybox
 *
 *
 * Copyright (C) 2000 by Matt Kraai <kraai@alumni.carnegiemellon.edu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <stdio.h>
#include <io.h>
//#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <windows.h>

struct suffix_mult {
	const char *suffix;
	int mult;
};

void perror_msg_and_die2(const char *s, char * arg1)
//void perror_msg_and_die2(const char *s, ...)
{
	fprintf(stderr,s, arg1); 
	perror(" ");
//	printf("hier");
	putc('\n', stderr);
	exit(EXIT_FAILURE);
}

size_t safe_read(int fd, void *buf, size_t count)
{
	size_t n;

	do {
		n = read(fd, buf, count);
	} while (n < 0);
//	} while (n < 0 && errno == EINTR);

	return n;
}
int full_write(int fd, const char *buf, int len)
{
	int cc;
	int total;

	total = 0;

	while (len > 0) {
		cc = write(fd, buf, len);

		if (cc < 0)
			return -1;

		buf += cc;
		total += cc;
		len -= cc;
	}

	return total;
}

unsigned long parse_number(const char *numstr,
		const struct suffix_mult *suffixes)
{
	const struct suffix_mult *sm;
	unsigned long int ret;
	int len;
	char *end;
	
	ret = strtoul(numstr, &end, 10);
	if (numstr == end)
		perror_msg_and_die2("invalid number `%s'", numstr);
	while (end[0] != '\0') {
		sm = suffixes;
		while ( sm != 0 ) {
			if(sm->suffix) {
				len = strlen(sm->suffix);
				if (strncmp(sm->suffix, end, len) == 0) {
					ret *= sm->mult;
					end += len;
					break;
				}
			sm++;
			
			} else
				sm = 0;
		}
		if (sm == 0)
			perror_msg_and_die2("invalid number `%s'", numstr);
	}
	return ret;
}


static const struct suffix_mult dd_suffixes[] = {
	{ "c", 1 },
	{ "w", 2 },
	{ "b", 512 },
	{ "kD", 1000 },
	{ "k", 1024 },
	{ "MD", 1000000 },
	{ "M", 1048576 },
	{ "GD", 1000000000 },
	{ "G", 1073741824 },
	{ NULL, 0 }
};

int main(int argc, char **argv)
{
	int i, ifd, ofd, oflag, sync_flag = FALSE, trunc = TRUE;
	size_t in_full = 0, in_part = 0, out_full = 0, out_part = 0;
	size_t bs = 512, count = -1;
	size_t n;
	off_t seek = 0, skip = 0;
	char *infile = NULL, *outfile = NULL, *buf;

	for (i = 1; i < argc; i++) {
		if (strncmp("bs=", argv[i], 3) == 0){
			bs = parse_number(argv[i]+3, dd_suffixes);
			continue;
		}
		if (strncmp("count=", argv[i], 6) == 0){
			count = parse_number(argv[i]+6, dd_suffixes);
			continue;
		}
		if (strncmp("seek=", argv[i], 5) == 0){
			seek = parse_number(argv[i]+5, dd_suffixes);
			continue;
		}
		if (strncmp("skip=", argv[i], 5) == 0){
			skip = parse_number(argv[i]+5, dd_suffixes);
			continue;
		}
		if (strncmp("if=", argv[i], 3) == 0){
			infile = argv[i]+3;
			continue;
		}
		if (strncmp("of=", argv[i], 3) == 0){
			outfile = argv[i]+3;
			continue;
		}
		if (strncmp("conv=", argv[i], 5) == 0) {
			buf = argv[i]+5;
			while (1) {
				if (strncmp("notrunc", buf, 7) == 0) {
					trunc = FALSE;
					buf += 7;
				} else if (strncmp("sync", buf, 4) == 0) {
					sync_flag = TRUE;
					buf += 4;
				} else {
					perror_msg_and_die2("invalid conversion `%s'", argv[i]+5);
				}
				if (buf[0] == '\0')
					break;
				if (buf[0] == ',')
					buf++;
			}
			continue;
		}
		
			printf("dd if=filename of=filename count=123 bs=512 ....\n");
			printf("\\\\.\\x: or \\\\.\\physicaldriveN with N=0,1,2.. gives access to\n");
			printf("partitions and disks (only nt-based windows, only in 512 byte blocks)");
		    exit(1);
//			show_usage();
	}

	buf = malloc(bs);

	if (infile != NULL) {
		if ((ifd = open(infile, O_RDONLY | _O_BINARY)) < 0)
			perror_msg_and_die2("%s", infile);
	} else {
//		ifd = STDIN_FILENO; //createfile CONIN$
//		if ((ifd = open("CONIN$", O_RDONLY)) < 0)
//			perror_msg_and_die2("%s", "CONIN$");
		ifd = 0;
		infile = "standard input";
	}

	if (outfile != NULL) {
		oflag = O_WRONLY | O_CREAT;

		if (!seek && trunc)
			oflag |= O_TRUNC;

//		if ((ofd = open(outfile, oflag, 0666)) < 0)
		if(NULL==strstr(outfile,"\\\\.\\")){
		 if ((ofd = open(outfile, oflag  | _O_BINARY, S_IREAD |S_IWRITE)) < 0) //file access
			perror_msg_and_die2("%s", outfile); //_S_IREAD | 
		}
		else{
		 if ((ofd = open(outfile, O_WRONLY | _O_BINARY, 0)) < 0) //partition or drive access
//		   if ((ofd = open(outfile, O_WRONLY | _O_BINARY, S_IREAD |S_IWRITE)) < 0) //partition or drive access
			perror_msg_and_die2("%s", outfile); //_S_IREAD |
		}

		if (seek && trunc) {
//			if (ftruncate(ofd, seek * bs) < 0)
//			if (lseek(ofd, seek * bs, SEEK_SET) < 0)
//				perror_msg_and_die2("%s", outfile);
		}
	} else {
//		ofd = STDOUT_FILENO;
//		if ((ofd = open("CONOUT$", O_WRONLY, 0)) < 0) //CreateFile stdout
//			perror_msg_and_die2("%s", "CONOUT$"); //_S_IREAD | 
		ofd=1;
		outfile = "standard output";
	}

	if (skip) {
		if (lseek(ifd, skip * bs, SEEK_CUR) < 0)
			perror_msg_and_die2("%s", infile);
	}

	if (seek) {
		if (lseek(ofd, seek * bs, SEEK_CUR) < 0)
			perror_msg_and_die2("%s", outfile);
	}

	while (in_full + in_part != count) {
		n = safe_read(ifd, buf, bs);
		if (n < 0)
			perror_msg_and_die2("%s", infile);
		if (n == 0)
			break;
		if (n == bs)
			in_full++;
		else
			in_part++;
		if (sync_flag) {
			memset(buf + n, '\0', bs - n);
			n = bs;
		}
		n = full_write(ofd, buf, n);
		if (n < 0)
			perror_msg_and_die2("%s", outfile);
		if (n == bs)
			out_full++;
		else
			out_part++;
	}

	fprintf(stderr, "%ld+%ld records in\n", (long)in_full, (long)in_part);
	fprintf(stderr, "%ld+%ld records out\n", (long)out_full, (long)out_part);

	return EXIT_SUCCESS;
}
